/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/analyze.c $
 $LastChangedRevision: 137 $
 $Date: 2014-07-15 16:38:33 -0400 (Tue, 15 Jul 2014) $
 $Author: bprommegger $
 ***********************************************************

 Previously part of SuperSNARK
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <strings.h>
#include "experimenter.h"
#include "getfom.h"
#include "stru_acc.h"
#include "hit_ratio.h"
#include "imagewise_roi.h"
#include "pointwse.h"
#include "fom_klds.h"
#include "fom_weighted_squared_distance.h"
#include "user_fom1.h"
#include "user_fom2.h"
#include "user_fom3.h"
#include "user_fom4.h"
#include "user_fom5.h"
#include "getiters.h"
#include "errorc.h"
#include "analyze.h"

void printout(FILE* fomfl, char* fom, int* itr1, int* itr2, int niters,	char* keywrd1, char* keywrd2, double* fom1, double* fom2);

/*--------------------------- analyze.c ---------------------------
 * This function reads the information in the file containing the comparisons
 * to be made and coordinates the calculation and storage of the appropriate
 * figures of merit (FOM).
 *
 * INPUTS:
 * exp - experiment number.
 * run - run number.
 * numpairs - for PAIRed structures: number of pairs of structures.
 * pairs - array of length 2*numpairs containing the indices of all paired
 * structures.
 * anaflname - name of file containing comparisons to be made.
 * fomfl - file pointer to the file which stores foms.
 *
 * OUTPUTS:
 * none.
 * ------------------------------------------------------------------------*/

void analyze(int exp, int run, int numpairs, int* pairs, char* anaflname, FILE* fomfl)
{
	int *itr1, *itr2, niters, stru, poin, hitr, usr1, usr2, usr3, usr4, usr5, iroi, klds, wsqd, num_mode = 1, compare = 0, mode_flag = 0;

	char string[MAXLINESIZE], key1[50], line[MAXLINESIZE], key2[50], keywrd1[50], keywrd2[50];

	FILE *anafl;

	strcpy(line, "In analyze.c: error in file");
	sprintf(line, "%s %s at line ", line, anaflname);

	if ((anafl = fopen(anaflname, "r")) == NULL)
	{
		errorc("in analyze.c: error in opening file ", anaflname);
	}

	fgets(string, sizeof(string), anafl); /* don't need first line */
	while (isSkip(string))     //jump over comment lines and blank lines
	{
		fgets(string, sizeof(string), anafl);
	}

	fgets(string, sizeof(string), anafl); /* second line is either a MODE * command or a COMPARE command */
	while (isSkip(string))     //jump over comment lines and blank lines
	{
		fgets(string, sizeof(string), anafl);
	}

	if (strncasecmp(string, "mode", 4) == 0)
	{ /* set mode_flag to 1 if */
		mode_flag = 1; /* second line is MODE */
	}
	else if (strncasecmp(string, "comp", 4) == 0)
	{
		;
	}
	else
	{
		errorc(line, string);
	}

	if (mode_flag == 1)
	{ /* if second line is MODE */
		fgets(string, sizeof(string), anafl); /* third line is COMPARE */
		while (isSkip(string))     //jump over comment lines and blank lines
		{
			fgets(string, sizeof(string), anafl);
		}
	}

	if (strncasecmp(string, "comp", 4) != 0)
	{
		errorc(line, string);
	}

	/* at this point string contains the COMPARE command */

	while (strncasecmp(string, "end", 3) != 0)
	{
		/* beginning of while loop */

		if (strncasecmp(string, "comp", 4) == 0)
		{
			/* If the string just read in has the COMPARE command:
			 (1) increment compare counter and read in the two keywords.
			 (2) get flags for the figures of merit (FOMs)
			 (3) get the iteration numbers to be compared
			 (4) compute the FOMs for these iteration numbers.
			 (5) free arrays for next set of iteration numbers. */

			compare++; /* keeps count of number of COMPAREs read in */
			sscanf(string, "%*s%s %s", key1, key2);
			strncpy(keywrd1, key1, 4);
			strncpy(keywrd2, key2, 4);
			keywrd1[4] = '\0';
			keywrd2[4] = '\0';
			getfom(string, &stru, &poin, &hitr, &usr1, &usr2, &usr3, &usr4,	&usr5, &iroi, &klds, &wsqd);

			getiters(anafl, &itr1, &itr2, &niters, string);

			fprintf(fomfl, " \n experiment= %d run= %d compare= %d\n", exp, run,
					compare);

			computefom(itr1, itr2, niters, numpairs, pairs, num_mode, keywrd1, keywrd2, stru, poin, hitr, usr1, usr2, usr3, usr4, usr5,	iroi, klds, wsqd, fomfl);

			free(itr1);
			free(itr2);
		}
		if (strncasecmp(string, "mode", 4) == 0)
		{
			/* If string just read in has the MODE command then increment
			 mode counter and read the next string (which has the COMPARE
			 command). */

			num_mode++;
			fgets(string, sizeof(string), anafl);
			while (isSkip(string))     //jump over comment lines and blank lines
			{
				fgets(string, sizeof(string), anafl);
			}
		}
	} /* end of while loop */

	fclose(anafl);
}

/*
 * This function coordinates the calculation and storage of the FOMs.
 *
 * INPUTS:
 * itr1 - array of length niters containing the iteration numbers to be
 * compared for the first algorithm.
 * itr2 - array of length niters containing the iteration numbers to be
 * compared for the second algorithm.
 * niters - number of iterations to be compared.
 * numpairs - for PAIRed structures: number of pairs of structures
 * pairs - array of length 2*numpairs containing the indices of all paired
 * structures.
 * num_mode - number of MODE commands read in so far.
 * keywrd1 - keyword which defines the first algorithm.
 * keywrd2 - keyword which defines the second algorithm.
 * stru - equals 1 if structural accuracy is to be computed, else equals 0.
 * poin - equals 1 if pointwise accuracy is to be computed, else equals 0.
 * hitr - equals 1 if hit-ratio is to be computed, else equals 0.
 * user - equals 1 if user-defined FOM is to be computed, else equals 0.
 * fomfl - file pointer to the file which stores the FOMs.
 *
 * OUTPUTS:
 * none
 */

void computefom(int* itr1, int* itr2, int niters, int numpairs, int* pairs, int num_mode, char* keywrd1, char* keywrd2, int stru, int poin, int hitr, int usr1, int usr2, int usr3, int usr4, int usr5, int iroi, int klds, int wsqd, FILE* fomfl)
{
	double *fom1, *fom2;

	/* allocate memory for the FOMs */
	fom1 = (double *) malloc(niters * sizeof(double));
	fom2 = (double *) malloc(niters * sizeof(double));

	if ((fom1 == NULL) || (fom2 == NULL))
	{
		errorc("in analyze.c: memory allocation failed in ", "computefom");
	}

	/* Depending on the values of the FOM flags compute and print the
	 corresponding FOM */

	if (stru == 1)
	{
		stru_acc(itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "Structural Accuracy", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}

	if (poin == 1)
	{
		pointwse(num_mode, itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "Point-wise Accuracy", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}

	if (usr1 == 1)
	{
		user_fom1(itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "User-defined-1   ", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}

	if (usr2 == 1)
	{
		user_fom2(itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "User-defined-2   ", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}

	if (usr3 == 1)
	{
		user_fom3(itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "User-defined-3   ", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}

	if (usr4 == 1)
	{
		user_fom4(itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "User-defined-4   ", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}

	if (usr5 == 1)
	{
		user_fom5(itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "User-defined-5   ", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}

	if (hitr == 1)
	{ /* hit-ratio is valid only if there
	 * exists paired structures */

		if (numpairs != 0)
		{
			hit_ratio(itr1, itr2, niters, pairs, numpairs, keywrd1, keywrd2, fom1, fom2);

			printout(fomfl, "Hit-ratio       ", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		}
		else
		{
			fprintf(fomfl, "\n    %s \t \t \t    %s\n", keywrd1, keywrd2);
			fprintf(fomfl,
					"***** No paired structure in phantom for this run\n");
		}
	}
	if (iroi == 1)
	{
		/* imagewise_roi is valid only if the number of paired structures is >= 2*/
		if (numpairs >= 2)
		{
			imagewise_roi(itr1, itr2, niters, pairs, numpairs, keywrd1, keywrd2,
					fom1, fom2);
			printout(fomfl, "Imagewise-ROI       ", itr1, itr2, niters, keywrd1,
					keywrd2, fom1, fom2);
		}
		else
		{
			fprintf(fomfl, "\n    %s \t \t \t    %s\n", keywrd1, keywrd2);
			fprintf(fomfl, "***** Not enough paired structure "
					"in phantom for this run\n");
			fprintf(fomfl, "******** The minimum number of paired structures "
					"should be >= 2 \n");
		}
	}
	if (klds == 1)
	{
		fom_klds(itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "Kullback-Leibler distance   ", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}
	if (wsqd == 1)
	{
		fom_weighted_squared_distance(itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
		printout(fomfl, "Weighted squared distance   ", itr1, itr2, niters, keywrd1, keywrd2, fom1, fom2);
	}


	free(fom1);
	free(fom2);
}

/*
 * This function prints out the computed FOMs
 *
 * INPUTS:
 * see the above routine.
 *
 * OUTPUTS:
 * none
 */

void printout(FILE* fomfl, char* fom, int* itr1, int* itr2, int niters, char* keywrd1, char* keywrd2, double* fom1, double* fom2)
{
	fprintf(fomfl, "\n    %s \t\t\t\t    %s\n iter \t %s \t\t iter \t %s\n", keywrd1, keywrd2, fom, fom);

	int i = 0;
	for (i = 0; i < niters; i++)
	{
		fprintf(fomfl, " %d \t %.25f\t %d\t %.25f\n", itr1[i], fom1[i], itr2[i], fom2[i]);
	}
}
