/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/analyze2.c $
 $LastChangedRevision: 137 $
 $Date: 2014-07-15 16:38:33 -0400 (Tue, 15 Jul 2014) $
 $Author: bprommegger $
 ***********************************************************

 Previously part of SuperSNARK
 */

#include <malloc.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include "experimenter.h"
#include "getfom.h"
#include "getiters.h"
#include "errorc.h"
#include "signif1.h"
#include "analyze2.h"

/*
 * This function reads the information in the file containing the comparisons
 * to be made and coordinates the calculation and storage of the significance
 * levels.
 *
 * INPUTS:
 * iexp - experiment number.
 * nrun - number of runs in a given experiment.
 * tot_PAIRS - number of phantoms with PAIRed structures used for a given
 * experiment.
 * anaflname - name of file containing comparisons to be made.
 *
 * OUTPUTS:
 * none.
 */
void analyze2(int iexp, int nrun, int tot_PAIRS, char* anaflname)
{
	int *itr1, *itr2, niters, stru, poin, hitr, usr1, usr2, usr3, usr4, usr5, iroi, klds, wsqd, compare = 0, mode_flag = 0;
	int i;

	char string[MAXLINESIZE], key1[50], sigflname[MAXLINESIZE], key2[50],
			keywrd1[50], keywrd2[50];

	float lowden = 0.0, highden = 0.0;

	FILE *anafl, *sigfl;

	if ((anafl = fopen(anaflname, "r")) == NULL)
	{
		errorc("in analyze2.c: error in opening file ", anaflname);
	}

	fgets(string, sizeof(string), anafl); // get name of file to store significance levels
	while (isSkip(string))     //jump over comment lines and blank lines
	{
		fgets(string, sizeof(string), anafl);
	}

	for (i = strlen(string) - 1; i >= 0; i--)
	{
		if (iscntrl(string[i]) || string[i] == ' ')
			string[i] = '\0';
	}
	if (strlen(string) == 0)
		errorc("in analyze2.c: error in opening file ", sigflname);

	strcpy(sigflname, string);
	sprintf(sigflname, "%s.%d", sigflname, iexp);

	if ((sigfl = fopen(sigflname, "w")) == NULL)
	{
		errorc("in analyze2.c: error in opening file ", sigflname);
	}

	fgets(string, sizeof(string), anafl); /* second line is either a MODE
	 * command or a COMPARE command */
	while (isSkip(string))     //jump over comment lines and blank lines
	{
		fgets(string, sizeof(string), anafl);
	}

	if (strncasecmp(string, "mode", 4) == 0)
	{
		/* set mode_flag to 1 if second line is MODE and read in
		 * MODE values */
		mode_flag = 1;
		sscanf(string, "%*s %f %f", &lowden, &highden);
	}

	if (mode_flag == 1)
	{ /* if second line is MODE */
		fgets(string, sizeof(string), anafl); /* third line is COMPARE */
		while (isSkip(string))     //jump over comment lines and blank lines
		{
			fgets(string, sizeof(string), anafl);
		}
	}

	/* at this point string contains the COMPARE command */

	while (strncasecmp(string, "end", 3) != 0)
	{
		/* beginning of while loop */

		if (strncasecmp(string, "comp", 4) == 0)
		{
			/* If the string just read in has the COMPARE command:
			 * (1) increment compare counter and read in the two keywords.
			 * (2) get flags for the figures of merit (FOMs)
			 * (3) get the iteration numbers to be compared
			 * (4) compute significance for these iteration numbers.
			 * (5) free arrays for next set of iteration numbers. */

			compare++; /* keeps count of number of COMPAREs read in */
			sscanf(string, "%*s%s %s", key1, key2);
			strncpy(keywrd1, key1, 4);
			strncpy(keywrd2, key2, 4);
			keywrd1[4] = '\0';
			keywrd2[4] = '\0';
			getfom(string, &stru, &poin, &hitr, &usr1, &usr2, &usr3, &usr4, &usr5, &iroi, &klds, &wsqd);
			getiters(anafl, &itr1, &itr2, &niters, string);
			computesig(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, keywrd1, keywrd2, stru, poin, hitr, iroi, klds, wsqd, usr1, usr2, usr3, usr4, usr5, lowden, highden);
			free(itr1);
			free(itr2);
		}

		if (strncasecmp(string, "mode", 4) == 0)
		{
			/* If string just read in has the MODE command then read in
			 * clip values and read the next string (which has the COMPARE
			 * command). */

			sscanf(string, "%*s %f %f", &lowden, &highden);
			fgets(string, sizeof(string), anafl);
			while (isSkip(string))     //jump over comment lines and blank lines
			{
				fgets(string, sizeof(string), anafl);
			}
		}
	} /* end of while loop */

	fclose(anafl);
	fclose(sigfl);
}

/*
 * This function coordinates the calculation and storage of the significance
 *
 * INPUTS:
 * iexp - experiment number.
 * nrun - number of runs for experiment iexp.
 * tot_PAIRS - number of phantoms with PAIRed structures for experiment iexp.
 * compare - number of COMPARE commands read in so far.
 * niters - number of iterations to be compared.
 * itr1 - array of length niters containing the iteration numbers to be
 * compared for the first algorithm.
 * itr2 - array of length niters containing the iteration numbers to be
 * compared for the second algorithm.
 * sigfl - file pointer to the file which stores the significance levels
 * keywrd1 - keyword which defines the first algorithm.
 * keywrd2 - keyword which defines the second algorithm.
 * stru - equals 1 if structural accuracy is to be computed, else equals 0.
 * poin - equals 1 if pointwise accuracy is to be computed, else equals 0.
 * hitr - equals 1 if hit-ratio is to be computed, else equals 0.
 * user - equals 1 if user-defined FOM is to be computed, else equals 0.
 * lowden - lower clip value of MODE command.
 * highden - higher clip value of MODE command.
 */

void computesig(int iexp, int nrun, int tot_PAIRS, int compare, int niters, int* itr1, int* itr2, FILE* sigfl, char* keywrd1, char* keywrd2, int stru, int poin, int hitr, int iroi, int klds, int wsqd, int usr1, int usr2, int usr3, int usr4, int usr5, float lowden, float highden)
{
	//  Depending on the values of the FOM flags compute and print the corresponding significance levels

	if (stru == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "Structural Accuracy", keywrd1, keywrd2, lowden, highden);
	}

	if (poin == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "Pointwise Accuracy", keywrd1, keywrd2, lowden, highden);
	}

	if (hitr == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "Hit-ratio", keywrd1, keywrd2, lowden, highden);
	}

	if (iroi == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "Imagewise-ROI", keywrd1, keywrd2, lowden, highden);
	}

	if (klds == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "Kullback-Leibler distance", keywrd1, keywrd2, lowden, highden);
	}

	if (wsqd == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "Weighted squared distance", keywrd1, keywrd2, lowden, highden);
	}

	if (usr1 == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "User-defined-1", keywrd1, keywrd2, lowden, highden);
	}

	if (usr2 == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "User-defined-2", keywrd1, keywrd2, lowden, highden);
	}

	if (usr3 == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "User-defined-3", keywrd1, keywrd2, lowden, highden);
	}

	if (usr4 == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "User-defined-4", keywrd1, keywrd2, lowden, highden);
	}

	if (usr5 == 1)
	{
		signif(iexp, nrun, tot_PAIRS, compare, niters, itr1, itr2, sigfl, "User-defined-5", keywrd1, keywrd2, lowden, highden);
	}
}
