/*
 **********************************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/creaph.cpp $
 $LastChangedRevision: 85 $
 $Date: 2014-07-02 16:07:08 -0400 (Wed, 02 Jul 2014) $
 $Author: agulati $
 **********************************************************************

 GENERATES A TEST PHANTOM
 *
 *				jk 11/18/2007 introducing variability
 *				jk 05/30/2008 redesigned and corrected implementation of variability
 *              the variability has to be added to each pixel at each energy
 *              level (polychromatic data) - in case of monochromatic data 
 *              there is simply only one energy level
 *              each pixel value is calculated according to the formula:
 *              1/(NAVE1^2) sum_{k=1 to NAVE1^2} sum_{j=1 to NOBJ} delta_{k,j}
 *                         sum_{i=1 to NERGY} (1+g_{j,i}) d_{j,i}p_{i}
 *              (see snark14 manual for explanation of the above value)
 */

#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <string.h>
#include <time.h>

#include "blkdta.h"
#include "creacm.h"
#include "geom.h"
#include "objects.h"
#include "uiod.h"

#include "file11.h"

#include "creaph.h"

#include "assert.h"

//jk 11/18/2007 introducing variability
//following header files needed for adding tissue variability 
#include "DIGGauss.h"
#include "consts.h"
//jk  05/30/08   a new header file needed
#include "spctrm.h"

void creaph()
{

	int i, j, k, n, ii, jj, index, ne; //loop counter variables
	REAL g; //temp variable for Gaussian value
	INTEGER rows; //temp variable for storing number of items in the rows

	REAL nmid;

	REAL unew;
	REAL vnew;
	REAL covru;
	REAL covrv;
	REAL sivru;
	REAL sivrv;
	REAL diam;
	REAL vbyusq;
	REAL xcntr;
	REAL ycntr;
	REAL denobj[7]; //temporarily stores the density of a pixel during computations
	//the value for each energy level is needed (max 7)

	REAL co;
	REAL si;

	INTEGER ny;
	REAL yo;
	INTEGER nx;
	REAL densit[7]; //temporarily stores the density of a pixel during computations
	//the value for each energy level is needed (max 7)

	INTEGER ny0;
	INTEGER nx0;

	INTEGER npts;
	INTEGER sample;
	INTEGER itype;

	REAL yc, xc, xo;

	// ALLOCATE WORKING SPACE - ONLY ONE ROW OF PIXELS IS NEEDED AT ONCE

	REAL * test_aray[8]; //stores a row of pixels at a time (alocated below)
	//each pixel needs a value for each energy level (7 of them)
	//the last item is used to store the average value
	REAL* coord_aray = NULL;

	//allocate enough space for one row (up to 7 energy levels + one more for average
	for (ne = 0; ne < 8; ne++)
		test_aray[ne] = new REAL[GeoPar.nelem];

	//jk 11/18/2007 introducing variability
	//allocate memory for the array that holds variability data
	//(for each of the energy levels)
	//do so only if the variability is not equal to zero
	memset(variab_array, 0, 7);
	if (variab_set)
	{
		for (ne = 0; ne < Spctrm.nergy; ne++)
		{
			variab_array[ne] = new REAL[GeoPar.nelem * GeoPar.nelem];
			memset(variab_array[ne], 0, 7);
		}

	}

	//jk 09/19/2008 modifying how polychromatic pseudo projections are computed
	//in the iterative runs of beam hardening correction
	memset(phantom_poly_array, 0, 7);
	for (ne = 0; ne < Spctrm.nergy; ne++)
	{
		phantom_poly_array[ne] = new REAL[GeoPar.nelem * GeoPar.nelem];
		memset(phantom_poly_array[ne], 0, GeoPar.nelem * GeoPar.nelem);
	}

	// THIS RUN IS FIRST GENERATION OF PHANTOM
	// ALLOCATE SPACE FOR SUBPIXELS
	npts = GeoPar.nelem * Creacm.nave1;

	coord_aray = new REAL[npts];

	nmid = (REAL) (npts / 2);

	// PRESET THE COORDINATES OF EACH SUBPIXEL CENTER
	for (n = 0; n < npts; n++)
	{
		coord_aray[n] = (((REAL) (n)) - nmid) * GeoPar.pixsiz
				/ ((REAL) (Creacm.nave1));
	}

	// COMPUTE THE NUMBER OF SAMPLES PER PIXEL FOR AVERAGING
	sample = Creacm.nave1 * Creacm.nave1;

	//set the seed for variability creation
	if (variab_set)
	{
		if (Creacm.varseed < 0)
			Srand(time(NULL));
		else if (Creacm.varseed > 0)
			Srand(Creacm.varseed);
		else
			Srand(1); //this generates the default sequence as if Srand was
		//not called (it is called here to start this default
		//sequence over
	}

	for (ne = 0; ne < Spctrm.nergy; ne++)
	{

		ny0 = -Creacm.nave1;
		nx0 = 0;

		// GET EACH PIXEL DENSITY
		for (j = 0; j < GeoPar.nelem; j++) 		// for each pix row
		{

			ny0 += Creacm.nave1;

			for (k = 0; k < GeoPar.nelem; k++)
			{
				// clear row buffer
				test_aray[ne][k] = 0;
			}

			// IF NO OBJECT SPECIFIED THEN GENERATE BLANK PHANTOM
			if (Creacm.nobj > 0)
			{

				// for each object
				for (index = 0; index < Creacm.nobj; index++)
				{

					itype = objects[index].type;

					// SKIP IF OBJECT WAS DELETED FROM LIST
					co = objects[index].cosang;
					si = objects[index].sinang;
					if (fabs(co - 1) < 0.000001)
						co = 1.;
					if (fabs(co + 1) < 0.000001)
						co = -1.;
					if (fabs(co) < 0.000001)
						co = 0.;
					if (fabs(si - 1) < 0.000001)
						si = 1.;
					if (fabs(si + 1) < 0.000001)
						si = -1.;
					if (fabs(si) < 0.000001)
						si = 0.;
					unew = objects[index].u;
					vnew = objects[index].v;
					covru = co / unew;
					covrv = co / vnew;
					sivru = si / unew;
					sivrv = si / vnew;
					// 9/12/2012 jklukowska					
					//if the unew and/or vnew are smaller than the size of the pixel,
					//use the size of the pixel as max 
					diam = (REAL) 1.5 * MAX0(MAX0(unew, vnew), GeoPar.pixsiz);
//                    diam = (REAL) 1.5 * MAX0(unew, vnew);
					vbyusq = (vnew / unew) * (vnew / unew);
					xcntr = objects[index].cx;
					ycntr = objects[index].cy;

					denobj[ne] = objects[index].den1[ne];

					ny = ny0;

					// for each subpix
					for (jj = 0; jj < Creacm.nave1; jj++)
					{

						// SHIFT THE ORIGIN TO THE CENTER OF THE OBJECT
						yo = -coord_aray[ny] - ycntr;
						ny++;

						// TEST IF THE POINT (X,Y) IS INSIDE THE INDEXED OBJECT
						// IF THE ROW WE ARE AT IS OUTSIDE MAXIMUM EXTENT OF
						// OBJECT, GET NEXT OBJECT
						if (fabs(yo) > diam)
							goto L95;
						// goto nex object

						// MOVE ACROSS COLUMNS

						nx = nx0;
						// for each pixel in row
						for (i = 0; i < GeoPar.nelem; i++)
						{
							densit[ne] = 0.0;

							// for each subpixel
							for (ii = 0; ii < Creacm.nave1; ii++)
							{

								xo = coord_aray[nx] - xcntr;
								nx++;

								// IF COLUMN IS LEFT OF EXTENT, INCREASE X
								if (xo < -diam)
									goto L70;
								// goto next column

								// IF COLUMN IS RIGHT OF EXTENT,UPDATE PIXEL ARRAY,THEN GET NEXT Y
								if (xo > diam)
								{
									// HERE xo .gt. diam
									test_aray[ne][i] += densit[ne];

									goto L90;
									// goto next row
								}
								// pixel within object radius
								// ROTATE BY ANGLE OF TILT
								xc = (REAL) fabs(xo * covru + yo * sivru);

								// CHECK XC FOR FUZZ
								if (fabs(xc) < 0.000001)
									xc = 0.0;
								if (fabs(xc + 1.0) < 0.000001)
									xc = -1.0;
								if (fabs(xc - 1.0) < 0.000001)
									xc = 1.0;

								if (xc <= 1.0)
								{

									yc = (yo * covrv - xo * sivrv);

									// CHECK YC FOR FUZZ
									if (fabs(yc) < 0.000001)
										yc = 0.0;
									if (fabs(yc - 1.0) < 0.000001)
										yc = 1.0;
									if (fabs(yc + 1.0) < 0.000001)
										yc = -1.0;

									if (yc <= 1.0)
									{
										switch (itype)
										{
										case SOT_elip: // ELLIPSE

											if ((yc >= -1.0)
													&& ((xc * xc + yc * yc - 1.0)
															< 0.000001))
											{
												// added tolerance - swr 11/26/04
												densit[ne] += denobj[ne];
											}
											break;

										case SOT_rect: // RECTANGLE
											if (yc >= -1.0)
											{
												densit[ne] += denobj[ne];
											}
											break;

										case SOT_tria: // TRIANGLE
											// CHECK TO SEE IF POINT IS CONTAINED IN TRIANGULAR AREA
											if ((yc >= 0.0)
													&& (xc <= (1.0 - yc)))
											{
												densit[ne] += denobj[ne];
												// NOW THE POINT IS INSIDE
												// THE OBJECTS BOUNDARIES
											}
											break;

										case SOT_segm: // SEGMENT
											if (yc <= 0.0)
											{

												// CHECK TO SEE IF POINT IS CONTAINED IN TRIANGULAR AREA
												if (xc <= (1.0 - yc))
												{
													//if(xc > 1.0-yc) goto L60;

													// TEST IF POINT IS BEYOND CIRCULAR ARC
													if ((xc * xc)
															<= (1.0
																	+ yc
																			* vbyusq
																			* (2.0
																					- yc)))
													{
														// NOW THE POINT IS INSIDE THE OBJECTS BOUNDARIES
														densit[ne] +=
																denobj[ne];

													}
												}
											}
											break;

										case SOT_sect: // SECTOR

											// CHECK TO SEE IF POINT IS CONTAINED IN TRIANGULAR AREA
											if (xc <= (1.0 - yc))
											{
												//if(xc > 1.0-yc) goto L60;
												// TEST IF POINT IS BEYOND CIRCULAR ARC
												if ((xc * xc)
														<= (1.0
																+ yc * vbyusq
																		* (2.0
																				- yc)))
												{
													// NOW THE POINT IS INSIDE THE OBJECTS BOUNDARIES
													densit[ne] += denobj[ne];
												}
											}
											break;

										}
									}
								}
								L70: ;
							} // for(ii
							test_aray[ne][i] += densit[ne];

						} //for(i
						L90: ;
					} //for(jj

					L95: ;
				} //for(index

			}

			rows = j * GeoPar.nelem; //this avoids repetitive multiplications below

			for (k = 0; k < GeoPar.nelem; k++)
			{
				// TAKE THE AVERAGE BY DIVIDING BY THE NUMBER OF SAMPLES
				if (Creacm.nave1 != 1)
				{
					test_aray[ne][k] /= sample;
				}
				//jk 11/18/2007 introducing variability
				//ADD TISSUE VARIABILITY TO THE PHANTOM DATA
				//variab_array[row][column] is assigned the value of variability for
				//each pixel

				//if the pixel belongs to the background (i.e. the density value
				//is zero for each energy level) the value is set to 0
				//(no variability is added to the background pixels)
				if (variab_set)
				{

					if (fabs(test_aray[ne][k]) > Consts.zero)
					{
						g = Gauss(0.0, variability) * test_aray[ne][k];
						test_aray[ne][k] += g;
						variab_array[ne][k + rows] = g;

					}

					else
					{
						variab_array[ne][k + rows] = 0;
					}

				}

				phantom_poly_array[ne][k + rows] = test_aray[ne][k];
			}
			for (k = 0; k < GeoPar.nelem; k++)
			{
				assert(phantom_poly_array[ne][rows + k] == test_aray[ne][k]);
			}

			//write row of pixels to file11 (use the weighted average
			//of all the energy levels)
			//File11.WriteRowOfPhanPixels(test_aray[7], GeoPar.neleM);
			// jk 07/08/2008 use the value of the first energy level
			// for phantom, instead of the weighted average)
			if (ne == 0)
				File11.WriteRowOfPhanPixels(test_aray[0], GeoPar.nelem);

		}  //for j
	} //for ne
	File11.WriteEndOfPhan();

	//delete allocated memory
	if (coord_aray != NULL)
		delete[] coord_aray;

	for (ne = 0; ne < 8; ne++)
	{
		if (test_aray[ne] != NULL)
			delete[] test_aray[ne];
	}
	//delete[] test_aray;

	return;
}
