/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/raylen.cpp $
 $LastChangedRevision: 91 $
 $Date: 2014-07-02 17:34:27 -0400 (Wed, 02 Jul 2014) $
 $Author: agulati $
 ***********************************************************

 RAYLEN RETURNS THE LENGTH OF THE RAY IN AN ELEMENTAL OBJECT

 SEE SNARK05 MANUAL FOR A FULL DESCRIPTION OF THE FUNCTIONING OF
 THE ROUTINE

 TYPE = 1 FOR ELLIPSE/CIRCLE
 TYPE = 2 FOR RECTANGLE/SQUARE
 TYPE = 3 FOR ISOS. TRIANGLE
 TYPE = 4 FOR SEGMENT OF CIRCLE
 TYPE = 5 FOR SECTOR OF CIRCLE
 AX,AY AND MX,MY DEFINE THE RAY:
 (AX,AY) COORDINATES OF ANY POINT ON THE RAY;
 (MX,MY) ARE THE X- AND Y- DIRECTION COSINES OF THE RAY
 IF THE RAY MAKES AN ANGLE THETA WITH X-AXIS THEN MX = COS(THETA)
 AND MY = SIN(THETA)

 (CX,CY) ARE THE COORDINATES OF THE CENTER OF THE OBJECT

 (U,V) ARE THE HALF- AXIS LENGTHS OF THE OBJECT

 (PX,PY) ARE THE COSINE AND SINE OF THE ANGLE OF ROTATION OF THE
 OBJECT ( THE ANGLE U MAKES WITH THE X-AXIS)
 */

#include <cstdlib>
#include <cstdio>
#include <cmath>

#include "blkdta.h"
#include "raylen.h"
#include "consts.h"

REAL raylen(SNARK_Object_type type, REAL ax, REAL ay, REAL mx, REAL my, REAL cx,
		REAL cy, REAL u, REAL v, REAL px, REAL py)
{
	REAL raylen_temp;
	REAL xo, yo;

	REAL xnew, ynew, unew, vnew, xdc, ydc;

	REAL diam;

	REAL temp, slope;
	REAL ycept;
	REAL xcept;

	REAL x1, x2;

	REAL a, b, c, d, bb;

	BOOLEAN rotate;

	REAL vtou, vsu;

	raylen_temp = 0.0;

	// SHIFT THE ORIGIN TO THE CENTER OF THE OBJECT

	xo = ax - cx;
	yo = ay - cy;

	// ROTATE ABOUT ORIGIN BY AN ANGLE = (U,X-AXIS) SO THAT U AND V ARE
	// ALONG THE NEW X AND Y AXES RESPECTIVELY

	xnew = xo * px + yo * py;
	ynew = -xo * py + yo * px;
	unew = u;
	vnew = v;
	xdc = mx * px + my * py;
	ydc = -mx * py + my * px;


	// FIND THE CIRCLE COMPLETELY ENCLOSING OBJECT

	diam = (REAL) 2.1 * MAX0(unew, vnew);

	// CHECK IF RAY HAS SLOPE .LE. 1.0  FOR COMPUTATIONAL PURPOSES
	// REMEMBER THIS FOR TRIA, SEGM, AND SECT SINCE THEY DO NOT
	// HAVE TWO-FOLD SYMMETRY

	rotate = FALSE;
	if (!((REAL) fabs(xdc) >= (REAL) fabs(ydc)))
	{

		// ROTATE ABOUT ORIGIN BY 90 DEG TO INSURE THAT SLOPE IS LE 1.0 IN
		// MAGNITUDE

		rotate = TRUE;
		temp = xnew;
		xnew = ynew;
		ynew = -temp;

		temp = xdc;
		xdc = ydc;
		ydc = -temp;

		temp = unew;
		unew = vnew;
		vnew = temp;
	}

	// SEE IF RAY IS ANYWHERE NEAR OBJECT CIRCLE

	slope = ydc / xdc;
	if ((REAL) fabs(slope) <= (REAL) 0.000001)
	{
		slope = 0;
	}

	ycept = ynew - slope * xnew;

	if ((REAL) fabs(ycept) < (REAL) 0.000001)
	{
		ycept = 0.0;
	}

	xcept = 0.0;

	if ((REAL) fabs(slope) > Consts.zero)
	{
		xcept = xnew - ynew / slope;
	}

	if (((REAL) fabs(xcept) > diam) && ((REAL) fabs(ycept) > diam))
	{
		return raylen_temp;
	}

	// if RECTANGLE or CIRCLE/ELLIPSE

	if (type == SOT_rect)
	{
		// CHECK IF RAY THROUGH RECTANGLE/SQUARE IS NEARLY PARALLEL TO X AXIS
		if (!((REAL) fabs(ydc) > (REAL) 0.0001))
		{
			if ((REAL) fabs(ynew) <= vnew)
				raylen_temp = (REAL) 2.0 * unew;

			return raylen_temp;
		}
		// GENERAL LINE INSIDE A RECTANGLE

#ifdef FFCOMPARE

		// there is still difference here between fortran
		raylen_temp = (REAL) fabs(
				(
						MAX0(-unew, MIN0(unew, ((REAL) (vnew - ynew) / slope) + xnew))
						- MAX0(-unew, MIN0(unew, ((REAL) -(vnew + ynew) / slope) + xnew))
				)
				/ xdc
		);
#else

		raylen_temp =
				(REAL) fabs(
						(
						MAX0(-unew, MIN0(unew, (vnew - ynew) / slope + xnew))
								- MAX0(-unew,
										MIN0(unew, -(vnew + ynew) / slope + xnew)))
								/ xdc);
#endif

		return raylen_temp;
	}

	// GENERAL LINE INSIDE ELLIPSE
	if (type == SOT_elip)
	{
		vsu = vnew * vnew + (unew * slope) * (unew * slope);
		d = vsu - ycept * ycept;

		// IF D POSITIVE, RAY INTERSECTS THE CIRCLE/ELLIPSE
		if (d > Consts.zero)
		{
			raylen_temp = (REAL) 2.0 * unew * vnew * (REAL) sqrt(d)
					/ (REAL) fabs(vsu * xdc);
#ifdef FFCOMPARE
			REAL zzz = (REAL) sqrt(d);
			REAL xxx = (REAL) 2.0 * unew * vnew * zzz;
			REAL yyy = (REAL) fabs(vsu * xdc);
			REAL aaa = (REAL) xxx / yyy;
			raylen_temp = (REAL) xxx / yyy;
#endif
		}
		return raylen_temp;
	}

	// TRIANGLE, SEGMENT, SECTOR

	if (!(type == SOT_segm))
	{

		// TRIANGLE AND TRIANGULAR PORTION OF SECTOR

		// SPECIAL CASE FOR TRIANGLE AND SECTOR
		// CHECK IF RAY IS NEARLY PARALLEL TO ONE OF THE SIDES

		vtou = vnew / unew;
		if (!((REAL) fabs(vtou - (REAL) fabs(slope)) >= (REAL) 0.0001))
		{
			if (!(rotate))
			{
				if (!((REAL) fabs(xcept) > unew))
				{
					if (slope >= 0.0)
						raylen_temp = (REAL) 0.5
								* (REAL) fabs((unew - xcept) / xdc);
					if (slope < 0.0)
						raylen_temp = (REAL) 0.5
								* (REAL) fabs((unew + xcept) / xdc);
				}
			}
			else
			{
				if (!((REAL) fabs(ycept) > vnew))
				{
					if (slope >= 0.0)
						raylen_temp = (REAL) 0.5
								* (REAL) fabs((vnew - ycept) / ydc);
					if (slope < 0.0)
						raylen_temp = (REAL) 0.5
								* (REAL) fabs((vnew + ycept) / ydc);
				}
			}
		}
		else
		{
			// TRIANGLE  GENERAL CASE

			if (!(rotate))
			{
				x1 = (vnew - ycept) / (slope + vtou);
				if ((x1 <= -Consts.zero) || (x1 > unew))  //bug 273, jklukowska
					x1 = xcept;
				x2 = (vnew - ycept) / (slope - vtou);
				if ((x2 < -unew) || (x2 >= Consts.zero))  //bug 273, jklukowska
					x2 = xcept;

				// RAY THROUGH VERTEX OF ISOSCELES TRIANGLE
				if (((REAL) fabs(x1) < (REAL) 0.00001)
						&& ((REAL) fabs(x2) < (REAL) 0.00001)
						&& ((REAL) fabs(xcept) <= unew))
					x1 = xcept;
				raylen_temp = (REAL) fabs((x1 - x2) / xdc);
			}
			else
			{

				x1 = (vnew - ycept) / (slope + vtou);
				if ((x1 < 0.0) || (x1 >= unew + 0.00001))
					x1 = 0.0;
				x2 = (-vnew - ycept) / (slope - vtou);
				if ((x2 < 0.0) || (x2 >= unew + 0.00001))
					x2 = 0.0;
				// RAY THROUGH VERTEX OF ISOSCELES TRIANGLE

				if (((REAL) fabs(x1 - unew) < (REAL) 0.000001)
						&& ((REAL) fabs(x2 - unew) < (REAL) 0.000001)
						&& ((REAL) fabs(ycept) <= vnew))
					x2 = 0.0;
				raylen_temp = (REAL) fabs((x1 - x2) / xdc);
			}

		}

		if (type == SOT_tria)
		{
			return raylen_temp;
		}
	}

	// SEGMENT OR THE SEGMENT PORTION OF THE SECTOR OF A CIRCLE

	if (!rotate)
	{
		if (((REAL) fabs(xcept) > unew) && (ycept > 0.0))
		{
			return raylen_temp;
		}

		a = (REAL) 1.0 + slope * slope;
		bb = ycept - vnew;
		b = slope * bb;
		c = bb * bb - (unew * unew + vnew * vnew);
		d = b * b - a * c;
		if (d <= 0.0)
		{
			return raylen_temp;
		}

		// THE RAY DOES INTERSECT THE CIRCLE
		d = (REAL) sqrt(d);
		if (b >= 0.0)
			x1 = -(b + d) / a;
		if (b < 0.0)
			x1 = (-b + d) / a;
		x2 = c / (a * x1);
		if (x1 * slope + ycept > 0.0)
			goto L111;
		if (x2 * slope + ycept > 0.0)
			goto L112;

		// X COMPONENT OF INTERSECTION LENGTH IS DIFFERENCE OF ROOTS OF QUADR
		if (raylen_temp > 0.0)
		{
			return raylen_temp;
		}
		raylen_temp += (REAL) 2.0 * d / (REAL) fabs(a * xdc);
		return raylen_temp;

		// RAY INTERSECTS LINEAR BOUNDARY OF SEGMENT
		L111: x1 = xcept;
		if (x2 * slope + ycept <= 0.0)
			goto L113;

		L112: x2 = xcept;

		L113: raylen_temp += (REAL) fabs((x1 - x2) / xdc);
		return raylen_temp;
	}

	if (((REAL) fabs(ycept) > vnew) && (xcept > 0.0))
	{
		return raylen_temp;
	}

	a = (REAL) 1.0 + slope * slope;
	b = slope * ycept - unew;
	c = ycept * ycept - vnew * vnew;
	d = b * b - a * c;

	if (d <= 0.0)
	{
		return raylen_temp;
	}

	d = (REAL) sqrt(d);
	if (b >= 0.0)
		x1 = -(b + d) / a;
	if (b < 0.0)
		x1 = (-b + d) / a;
	x2 = c / (a * x1);
	if (x1 > 0.0)
		goto L121;
	if (x2 > 0.0)
		goto L122;

	// X COMPONENT OF INTERSECTION LENGTH IS DIFFERENCE OF ROOTS OF QUADR
	if (raylen_temp > 0.0)
	{
		return raylen_temp;
	}
	raylen_temp += (REAL) 2.0 * d / (REAL) fabs(a * xdc);
	return raylen_temp;

	// RAY INTERSECTS LINEAR BOUNDARY OF SEGMENT
	L121: x1 = 0.0;
	if (x2 <= 0.0)
		goto L123;

	L122: x2 = 0.0;

	L123: raylen_temp += (REAL) fabs((x1 - x2) / xdc);
	return raylen_temp;
}
